/*
 * PSEmu Pro Plugin linux Sound routines.
 *
 * By: linuzappz <linuzappz@hotmail.com>
 *
 * Needs the pthread library, it seems to work fine, send any comments or bugs
 * to my email.
 * Lacks reverberation, adsr, noise and fm mode.
 *
 * Thanks to Pete for his help.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <gtk/gtk.h>

typedef char HWND;

#include "PSEmu Plugin Defs.h"
#include "Decode XA.h"

static char *LibName = "LinuzAppz Spu Driver";

const unsigned char version = 1; // PSEmu 1.x library
const unsigned char revision = 1;
const unsigned char build = 6;

#define SNDDEV		"/dev/dsp"
char Stereo = 1;	// 1 - Stereo : 0 - Mono

int _Thread = 0;

#define CreateThread \
	    if (_Thread == 0) { _Thread = 1; pthread_create(&thread, NULL, SndThread, NULL); }

#define ExitThread \
	    { _Thread = 0; pthread_exit(NULL); }

int spu_sbaddr;
short spureg[(0x1e00-0x1c00)/2];
short *spumem;
int SndDev = -1;
pthread_t thread = -1;
unsigned short VolL = 0x3fff, VolR = 0x3fff;
unsigned short CdVolL = 0x7fff, CdVolR = 0x7fff;
int SpuOn = 0, Muted = 0, DmaMode = 0;
//void (__stdcall *irqCallback)(void) = 0;

void VoiceOn(unsigned int chn);
void VoiceOff(unsigned int chn);
int VoiceStatus(int first, int last);
void *SndThread(void *arg);
void LoadConf();
void SaveConf();
int vag_depack(unsigned short *entree, short *sortie);

typedef struct _Voice_ {
    short volL,volR;		// Left and Right Volume
    unsigned short freq;	// Voice Pitch
    unsigned short saddr;	// Start Adrress of Sound
    unsigned short ads;		// Attack/Decay/Sustain Level
    unsigned short sr;		// Sustain/Release Rate
    unsigned short adsrV;	// ADSR Volume
    unsigned short raddr;	// Repeat Address
    int on;			// 1 == On, 0 == Off
    int pos;			// Current Position
    int size;			// Size of Sound
    short *buffer;		// Voice Buffer (65536 * 2 bytes)
    unsigned int count;		// Count to 44100 from Voice Pitch
} StVoice;

StVoice Voice[25];

char *PSEgetLibName(void) {
    return LibName;
}

unsigned long PSEgetLibType(void) {
    return PSE_LT_SPU;
}

unsigned long PSEgetLibVersion(void) {
    return version<<16|revision<<8|build;
}

long SPUinit(void) {
    int i;

    spumem = (short *)malloc(512*1024);
    if (spumem == NULL) return -1;

    for (i=0;i<25;i++) memset(&Voice[i],0,sizeof(Voice[i]));

    for (i=0;i<25;i++)
	Voice[i].buffer = (short *) malloc (65536*sizeof(short));

    LoadConf();

    return 0;
}

long SPUshutdown(void) {
    int i;

    if (spumem != NULL) {
	free(spumem);
	spumem = NULL;
    }
    for (i=0; i<25; i++)
        if (Voice[i].buffer != NULL) {
	    free(Voice[i].buffer);
	    Voice[i].buffer = NULL;
	}

    return 0;
}

long SPUopen(void) {
    int i;

    if (SndDev != -1) return 0; // already opened

    // Open Driver

    SndDev = open(SNDDEV, O_WRONLY);
    if (SndDev == -1) return -1;

    if (ioctl(SndDev,SNDCTL_DSP_RESET)==-1) return -1;

    i = AFMT_S16_LE;
    if (ioctl(SndDev,SNDCTL_DSP_SETFMT,&i)==-1) { close(SndDev); SndDev = -1; return -1; }

    i = 44100;
    ioctl (SndDev,SNDCTL_DSP_SPEED,&i);
    i = 1 + Stereo;
    ioctl (SndDev,SNDCTL_DSP_CHANNELS,&i);

    return 0;
}

long SPUclose(void) {
    if (SndDev != -1) { close (SndDev); SndDev = -1; }
    if (thread != -1) { pthread_cancel(thread); thread = -1; }
    return 0;
}

// New Interface

void SPUwriteRegister(unsigned long reg, unsigned short val) {
    spureg[(reg-0x1f801c00)/2] = val;
//    printf ("write Register %x: %x\n", reg, val);
    if ((reg>=0x1f801c00) & (reg<0x1f801d80)) {
	int n = (reg>>4)&0x1f;

	switch (reg&0xf) {
	    case 0x0: 
		if (val&0x8000) { // sweep mode
		    Voice[n].volL = val<<7;
		}
		else { // volume mode
		    Voice[n].volL = val;
		}
		break;
	    case 0x2:
		if (val&0x8000) { // sweep mode
		    Voice[n].volR = val<<7;
		}
		else { // volume mode
		    Voice[n].volR = val;
		}
		break;
	    case 0x4: Voice[n].freq = val * 44100 / 4096; break;
	    case 0x6: if (val <= 0x7ffff) Voice[n].saddr = val * 8; 
		      else Voice[n].saddr = 0;
		      break;
	    case 0x8: Voice[n].ads = val; break;
	    case 0xa: Voice[n].sr = val; break;
	    case 0xc: Voice[n].adsrV = val; break;
	    case 0xe: Voice[n].raddr = val * 8; break;
	}
	return;
    }
    switch(reg) {
    case 0x1f801d80: // volume left
	if (val&0x8000) { // sweep mode
	    VolL = val<<7;
	}
	else { // volume mode
	    VolL = val;
	}
	break;
    case 0x1f801d82: // volume right
	if (val&0x8000) { // sweep mode
	    VolR = val<<7; 
	}
	else { // volume mode
	    VolR = val; 
	}
	break;
    case 0x1f801d84: // reverberation depth left
	break;
    case 0x1f801d86: // reverberation depth right
	break;
    case 0x1f801d88: // voice on 0 - 15
	if (val) VoiceOn(val); break;
    case 0x1f801d8a: // voice on 16 - 23
	if (val) VoiceOn(val<<16); break;
    case 0x1f801d8c: // voice off 0 - 15
	if (val) VoiceOff(val); break;
    case 0x1f801d8e: // voice off 16 - 23
	if (val) VoiceOff(val<<16); break;
    case 0x1f801d90: // channel fm mode 0 - 15
	break;
    case 0x1f801d92: // channel fm mode 16 - 23
	break;
    case 0x1f801d94: // channel noise mode 0 - 15
	break;
    case 0x1f801d96: // channel noise mode 16 - 23
	break;
    case 0x1f801d98: // channel reverb mode 0 - 15
	break;
    case 0x1f801d9a: // channel reverb mode 16 - 23
	break;
    case 0x1f801da2: // reverb work area start
	break;
    case 0x1f801da6: // spu sbaddr
        spu_sbaddr = val * 8;
	if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
        break;
    case 0x1f801da8: // spu data
	if (DmaMode == 0x10) { // non dma mode
	    spumem[spu_sbaddr/2] = (short)val;
	    spu_sbaddr+=2;
	    if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
	}
        break;
    case 0x1f801daa: // spu control
	SpuOn = val&0x8000;
	Muted = val&0x4000;
	DmaMode = val&0x30;
        break;
    case 0x1f801db0: // cd volume left
	CdVolL = val&0x7fff;
	break;
    case 0x1f801db2: // cd volume right
	CdVolR = val&0x7fff;
	break;
    }
}

unsigned short SPUreadRegister(unsigned long reg) {
//    printf ("write Register %x\n", reg);
    switch (reg){
	case 0x1f801d9c: // channel on/off 0 - 15
	    return VoiceStatus(0, 16);
	case 0x1f801d9e: // channel on/off 16 - 23
	    return VoiceStatus(16, 24);
	case 0x1f801da6: // spu sbaddr
    	    return spu_sbaddr / 8;
	case 0x1f801da8: // spu data
	    if (DmaMode == 0x10) { // non dma mode
		int ret = spumem[spu_sbaddr/2];
		spu_sbaddr+=2;
		if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
		return ret;
	    }
	    return 0;
	default:
	    return spureg[(reg-0x1f801c00)/2];
    }
    return 0;
}

unsigned short SPUreadDMA(void) {
    int ret = spumem[spu_sbaddr/2];
    spu_sbaddr+=2;
    if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
    return ret;
}

void SPUwriteDMA(unsigned short val) {
    spumem[spu_sbaddr/2] = (short)val;
    spu_sbaddr+=2;
    if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
}

void SPUwriteDMAMem(unsigned short *pMem, int iSize) {
    while (iSize > 0) {
	spumem[spu_sbaddr/2] = (short)*pMem;
	spu_sbaddr+=2;
	if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
	iSize--;
	pMem++;
    }
}

void SPUreadDMAMem(unsigned short *pMem, int iSize) {
    while (iSize > 0) {
	*pMem = spumem[spu_sbaddr/2];
	spu_sbaddr+=2;
	if (spu_sbaddr > 0x7ffff) spu_sbaddr = 0;
	iSize--;
	pMem++;
    }
}

void SPUplayADPCMchannel(xa_decode_t *xap) {
    memcpy(Voice[24].buffer, xap->pcm, xap->nsamples*2*2);
    Voice[24].size = xap->nsamples*2;
    Voice[24].on = 1;
    Voice[24].count = 0;
    Voice[24].pos = 0;
    Voice[24].raddr = 0;
    Voice[24].freq = xap->freq;
    Voice[24].volL = CdVolL>>1;
    Voice[24].volR = CdVolR>>1;
    CreateThread;
}

// Old Interface

unsigned short SPUgetOne(unsigned long val) {
    if (val > 0x7ffff) return 0;
    return spumem[val/2];
}

void SPUputOne(unsigned long val, unsigned short data) {
    if (val > 0x7ffff) return;
    spumem[val/2] = data;
}

void SPUsetAddr(unsigned char ch, unsigned short waddr) {
    Voice[ch].saddr = waddr * 8; 
}

void SPUsetPitch(unsigned char ch, unsigned short pitch) {
    Voice[ch].freq = pitch * 44100 / 4096;
}

void SPUsetVolumeL(unsigned char ch, short vol) {
    Voice[ch].volL = vol;
}

void SPUsetVolumeR(unsigned char ch, short vol) {
    Voice[ch].volR = vol;
}

void SPUstartChannels1(unsigned short channels) {
    SpuOn = 1; Muted = 1;
    VoiceOn(channels);
}

void SPUstartChannels2(unsigned short channels) {
    SpuOn = 1; Muted = 1;
    VoiceOn(channels<<16);
}

void SPUstopChannels1(unsigned short channels) {
    SpuOn = 1;
    VoiceOff(channels);
}

void SPUstopChannels2(unsigned short channels) {
    SpuOn = 1;
    VoiceOff(channels<<16);
}

/*void SPUregisterCallback(void (__stdcall *callback)(void)) {
    irqCallback = callback;
}*/

long SPUtest(void) {
    int dev;
    if ((dev = open(SNDDEV, O_WRONLY)) == -1) return -1;
    close(dev);
    return 0;
}

GtkWidget *ConfDlg;

void OnOk() {
    gtk_widget_hide(ConfDlg);
    SaveConf();
}

void OnCancel() {
    gtk_widget_hide(ConfDlg);
}

void OnCheck() {
    Stereo = 1 - Stereo;
}

long SPUconfigure(void) {
    GtkWidget *Check;
    GtkWidget *Ok,*Cancel;
    GtkWidget *Box,*Box1,*BBox;

    LoadConf();

    ConfDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(ConfDlg, 260, 80);
    gtk_window_set_title(GTK_WINDOW(ConfDlg), "SPU Config Dialog");
    gtk_window_set_position(GTK_WINDOW(ConfDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(ConfDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(ConfDlg), Box);
    gtk_widget_show(Box);

    Box1 = gtk_hbox_new(0, 0);
    gtk_box_pack_start(GTK_BOX(Box), Box1, FALSE, FALSE, 0);
    gtk_widget_show(Box1);

    Check = gtk_check_button_new_with_label("Use Stereo Sound");
    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(Check), Stereo);
    gtk_box_pack_start(GTK_BOX(Box1), Check, FALSE, FALSE, 0);
    gtk_widget_show(Check);
    gtk_signal_connect(GTK_OBJECT(Check), "clicked", GTK_SIGNAL_FUNC(OnCheck), NULL);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    Cancel = gtk_button_new_with_label("Cancel");
    gtk_signal_connect(GTK_OBJECT(Cancel), "clicked", GTK_SIGNAL_FUNC(OnCancel), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Cancel);
    GTK_WIDGET_SET_FLAGS(Cancel, GTK_CAN_DEFAULT);
    gtk_widget_show(Cancel);

    gtk_widget_show(ConfDlg);

    return 0;
}

GtkWidget *AboutDlg;

void OnAboutOk(GtkMenuItem *menuitem, gpointer userdata) {
    gtk_widget_hide(AboutDlg);
}

void SPUabout(void) {
    GtkWidget *Label;
    GtkWidget *Ok;
    GtkWidget *Box,*BBox;
    char AboutText[255];

    sprintf (AboutText,"%s %d.%d\n",LibName, revision, build);

    AboutDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(AboutDlg, 260, 80);
    gtk_window_set_title(GTK_WINDOW(AboutDlg), "SPU About Dialog");
    gtk_window_set_position(GTK_WINDOW(AboutDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(AboutDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(AboutDlg), Box);
    gtk_widget_show(Box);

    Label = gtk_label_new(AboutText);
    gtk_box_pack_start(GTK_BOX(Box), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnAboutOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    gtk_widget_show(AboutDlg);
}

void VoiceOn(unsigned int chn) {
    int i;

    if (!SpuOn) return;

    for (i=0; i<24; i++) {
	if (chn&(1<<i)) {
	    Voice[i].size = vag_depack((unsigned short*)(spumem+Voice[i].saddr/2),Voice[i].buffer);
	    Voice[i].on = 1;
	    Voice[i].count = 0;
	    Voice[i].pos = 0;
	    Voice[i].raddr = 0;
	    CreateThread;
	}
    }
}

void VoiceOff(unsigned int chn) {
    int i;

    if (!SpuOn) return;

    for (i=0; i<24; i++) {
	if (chn&(1<<i)) {
	    Voice[i].on = 0;
	}
    }
}

int VoiceStatus(int first, int last) {
    int ret = 0;

    if (!SpuOn) return 0;

    for (; first<last; first++) {
	ret |= Voice[first].on<<first;
    }
    return ret;
}

#define SIZE_SBUF	1024
short sbuf2[SIZE_SBUF*2];
int sbuf[SIZE_SBUF*2];

void *SndThread(void *arg) {
    int i,j,n;
    StVoice *chn;

    for (;;) {
	if ((!SpuOn) && (!Muted)) continue;
	n = 0;
	for (i=0; i<25; i++) {
	    chn = &Voice[i];
	    if (!chn->on) continue;
	    n++;
	    for (j=0; j<SIZE_SBUF; j++) {
		chn->count += chn->freq;
		if (chn->count >= 44100) {
		    chn->pos += (chn->count / 44100);
		    chn->count %= 44100;
		}

		if (Stereo) {
		    sbuf[j*2] += chn->buffer[chn->pos] * ((float)chn->volL/0x3fff);
		    sbuf[j*2+1] += -chn->buffer[chn->pos] * ((float)chn->volR/0x3fff);
		}
		else {
		    sbuf[j] += chn->buffer[chn->pos] * ((float)((chn->volL+chn->volR) / 2)/0x3fff);
		}

		if (chn->pos >= chn->size) {
		    if (chn->raddr > 0) { // repeat sound
			chn->count = 0;
			chn->pos = chn->raddr / 2;
		    }
		    else {
			VoiceOff(1<<i);
		    }
		    break;
		}
	    }
	}
	if (n>0) {
	    if (Stereo) {
		for (i=0; i<(SIZE_SBUF * 2); i+=2) {
		    sbuf2[i] = (sbuf[i] / n) * ((float)VolL/0x3fff);
		    sbuf2[i+1] = -(sbuf[i+1] / n) * ((float)VolR/0x3fff);
		}
	        write(SndDev, sbuf2, sizeof(sbuf2));
	    }
	    else {
		for (i=0; i<SIZE_SBUF; i++) {
		    sbuf2[i] = (sbuf[i] / n) * ((float)((VolL+VolR) / 2)/0x3fff);
		}
	        write(SndDev, sbuf2, sizeof(sbuf2)/2);
	    }
	    memset(sbuf, 0, sizeof(sbuf));
	}
	else {
	    ExitThread;
	}
    }
}

void LoadConf() {
    FILE *f;
    char cfg[256];

    sprintf (cfg, "%s/spu.cfg", getenv("HOME"));
    f = fopen (cfg, "r");
    if (f == NULL) return;
    fread(&Stereo, 1, 1, f);
    fclose(f);
}

void SaveConf() {
    FILE *f;
    char cfg[256];

    sprintf (cfg, "%s/spu.cfg", getenv("HOME"));
    f = fopen (cfg, "w");
    if (f == NULL) return;
    fwrite(&Stereo, 1, 1, f);
    fclose(f);
}